@@ -1,6 +1,6 @@
 /****************************************************************************
 **
-** Copyright (C) 2016 The Qt Company Ltd.
+** Copyright (C) 2017 The Qt Company Ltd.
 ** Contact: http://www.qt.io/licensing/
 **
 ** This file is part of the Qt Quick Templates 2 module of the Qt Toolkit.
@@ -92,6 +92,13 @@ static const int AUTO_REPEAT_INTERVAL = 100;
     This signal is emitted when the button is interactively clicked by the user.
 */
 
+/*!
+    \since QtQuick.Controls 2.2 (Qt 5.9)
+    \qmlsignal QtQuick.Controls::AbstractButton::toggled()
+
+    This signal is emitted when a checkable button is interactively toggled by the user.
+*/
+
 /*!
     \qmlsignal QtQuick.Controls::AbstractButton::pressAndHold()
 
@@ -104,13 +111,93 @@ static const int AUTO_REPEAT_INTERVAL = 100;
     This signal is emitted when the button is interactively double clicked by the user.
 */
 
-QQuickAbstractButtonPrivate::QQuickAbstractButtonPrivate() :
-    down(false), explicitDown(false), pressed(false), keepPressed(false), checked(false), checkable(false),
-    autoExclusive(false), autoRepeat(false), wasHeld(false),
-    holdTimer(0), delayTimer(0), repeatTimer(0), repeatButton(Qt::NoButton), indicator(nullptr), group(nullptr)
+QQuickAbstractButtonPrivate::QQuickAbstractButtonPrivate()
+    : down(false),
+      explicitDown(false),
+      pressed(false),
+      keepPressed(false),
+      checked(false),
+      checkable(false),
+      autoExclusive(false),
+      autoRepeat(false),
+      wasHeld(false),
+      holdTimer(0),
+      delayTimer(0),
+      repeatTimer(0),
+      pressButtons(Qt::NoButton),
+      indicator(nullptr),
+      group(nullptr)
 {
 }
 
+void QQuickAbstractButtonPrivate::handlePress(const QPointF &point)
+{
+    Q_Q(QQuickAbstractButton);
+    QQuickControlPrivate::handlePress(point);
+    pressPoint = point;
+    q->setPressed(true);
+
+    emit q->pressed();
+
+    if (autoRepeat)
+        startRepeatDelay();
+    else if (touchId != -1 || Qt::LeftButton == (pressButtons & Qt::LeftButton))
+        startPressAndHold();
+    else
+        stopPressAndHold();
+}
+
+void QQuickAbstractButtonPrivate::handleMove(const QPointF &point)
+{
+    Q_Q(QQuickAbstractButton);
+    QQuickControlPrivate::handleMove(point);
+    q->setPressed(keepPressed || q->contains(point));
+
+    if (!pressed && autoRepeat)
+        stopPressRepeat();
+    else if (holdTimer > 0 && (!pressed || QLineF(pressPoint, point).length() > QGuiApplication::styleHints()->startDragDistance()))
+        stopPressAndHold();
+}
+
+void QQuickAbstractButtonPrivate::handleRelease(const QPointF &point)
+{
+    Q_Q(QQuickAbstractButton);
+    QQuickControlPrivate::handleRelease(point);
+    bool wasPressed = pressed;
+    q->setPressed(false);
+    pressButtons = Qt::NoButton;
+
+    if (!wasHeld && (keepPressed || q->contains(point)))
+        q->nextCheckState();
+
+    if (wasPressed) {
+        emit q->released();
+        if (!wasHeld)
+            emit q->clicked();
+    } else {
+        emit q->canceled();
+    }
+
+    if (autoRepeat)
+        stopPressRepeat();
+    else
+        stopPressAndHold();
+}
+
+void QQuickAbstractButtonPrivate::handleUngrab()
+{
+    Q_Q(QQuickAbstractButton);
+    QQuickControlPrivate::handleUngrab();
+    pressButtons = Qt::NoButton;
+    if (!pressed)
+        return;
+
+    q->setPressed(false);
+    stopPressRepeat();
+    stopPressAndHold();
+    emit q->canceled();
+}
+
 bool QQuickAbstractButtonPrivate::isPressAndHoldConnected()
 {
     Q_Q(QQuickAbstractButton);
@@ -162,6 +249,15 @@ void QQuickAbstractButtonPrivate::stopPressRepeat()
     }
 }
 
+void QQuickAbstractButtonPrivate::toggle(bool value)
+{
+    Q_Q(QQuickAbstractButton);
+    const bool wasChecked = checked;
+    q->setChecked(value);
+    if (wasChecked != checked)
+        emit q->toggled();
+}
+
 QQuickAbstractButton *QQuickAbstractButtonPrivate::findCheckedButton() const
 {
     Q_Q(const QQuickAbstractButton);
@@ -207,20 +303,26 @@ QList<QQuickAbstractButton *> QQuickAbstractButtonPrivate::findExclusiveButtons(
     return buttons;
 }
 
-QQuickAbstractButton::QQuickAbstractButton(QQuickItem *parent) :
-    QQuickControl(*(new QQuickAbstractButtonPrivate), parent)
+QQuickAbstractButton::QQuickAbstractButton(QQuickItem *parent)
+    : QQuickControl(*(new QQuickAbstractButtonPrivate), parent)
 {
     setActiveFocusOnTab(true);
     setFocusPolicy(Qt::StrongFocus);
     setAcceptedMouseButtons(Qt::LeftButton);
+#if QT_CONFIG(cursor)
+    setCursor(Qt::ArrowCursor);
+#endif
 }
 
-QQuickAbstractButton::QQuickAbstractButton(QQuickAbstractButtonPrivate &dd, QQuickItem *parent) :
-    QQuickControl(dd, parent)
+QQuickAbstractButton::QQuickAbstractButton(QQuickAbstractButtonPrivate &dd, QQuickItem *parent)
+    : QQuickControl(dd, parent)
 {
     setActiveFocusOnTab(true);
     setFocusPolicy(Qt::StrongFocus);
     setAcceptedMouseButtons(Qt::LeftButton);
+#if QT_CONFIG(cursor)
+    setCursor(Qt::ArrowCursor);
+#endif
 }
 
 QQuickAbstractButton::~QQuickAbstractButton()
@@ -254,6 +356,7 @@ void QQuickAbstractButton::setText(const QString &text)
 
     d->text = text;
     setAccessibleName(text);
+    buttonChange(ButtonTextChange);
     emit textChanged();
 }
 
@@ -319,6 +422,7 @@ void QQuickAbstractButton::setPressed(bool isPressed)
     d->pressed = isPressed;
     setAccessibleProperty("pressed", isPressed);
     emit pressedChanged();
+    buttonChange(ButtonPressedChanged);
 
     if (!d->explicitDown) {
         setDown(d->pressed);
@@ -350,7 +454,7 @@ void QQuickAbstractButton::setChecked(bool checked)
 
     d->checked = checked;
     setAccessibleProperty("checked", checked);
-    checkStateSet();
+    buttonChange(ButtonCheckedChange);
     emit checkedChanged();
 }
 
@@ -383,7 +487,7 @@ void QQuickAbstractButton::setCheckable(bool checkable)
 
     d->checkable = checkable;
     setAccessibleProperty("checkable", checkable);
-    checkableChange();
+    buttonChange(ButtonCheckableChange);
     emit checkableChanged();
 }
 
@@ -431,7 +535,7 @@ void QQuickAbstractButton::setAutoRepeat(bool repeat)
 
     d->stopPressRepeat();
     d->autoRepeat = repeat;
-    autoRepeatChange();
+    buttonChange(ButtonAutoRepeatChange);
 }
 
 /*!
@@ -451,7 +555,7 @@ void QQuickAbstractButton::setIndicator(QQuickItem *indicator)
     if (d->indicator == indicator)
         return;
 
-    d->deleteDelegate(d->indicator);
+    QQuickControlPrivate::destroyDelegate(d->indicator, this);
     d->indicator = indicator;
     if (indicator) {
         if (!indicator->parentItem())
@@ -476,10 +580,8 @@ void QQuickAbstractButton::focusOutEvent(QFocusEvent *event)
 {
     Q_D(QQuickAbstractButton);
     QQuickControl::focusOutEvent(event);
-    if (d->pressed) {
-        setPressed(false);
-        emit canceled();
-    }
+    if (d->touchId == -1) // don't ungrab on multi-touch if another control gets focused
+        d->handleUngrab();
 }
 
 void QQuickAbstractButton::keyPressEvent(QKeyEvent *event)
@@ -490,12 +592,11 @@ void QQuickAbstractButton::keyPressEvent(QKeyEvent *event)
         d->pressPoint = QPoint(qRound(width() / 2), qRound(height() / 2));
         setPressed(true);
 
-        if (d->autoRepeat) {
+        if (d->autoRepeat)
             d->startRepeatDelay();
-            d->repeatButton = Qt::NoButton;
-        }
 
         emit pressed();
+        event->accept();
     }
 }
 
@@ -512,62 +613,15 @@ void QQuickAbstractButton::keyReleaseEvent(QKeyEvent *event)
 
         if (d->autoRepeat)
             d->stopPressRepeat();
+        event->accept();
     }
 }
 
 void QQuickAbstractButton::mousePressEvent(QMouseEvent *event)
 {
     Q_D(QQuickAbstractButton);
+    d->pressButtons = event->buttons();
     QQuickControl::mousePressEvent(event);
-    d->pressPoint = event->pos();
-    setPressed(true);
-
-    emit pressed();
-
-    if (d->autoRepeat) {
-        d->startRepeatDelay();
-        d->repeatButton = event->button();
-    } else if (Qt::LeftButton == (event->buttons() & Qt::LeftButton)) {
-        d->startPressAndHold();
-    } else {
-        d->stopPressAndHold();
-    }
-}
-
-void QQuickAbstractButton::mouseMoveEvent(QMouseEvent *event)
-{
-    Q_D(QQuickAbstractButton);
-    QQuickControl::mouseMoveEvent(event);
-    setPressed(d->keepPressed || contains(event->pos()));
-
-    if (!d->pressed && d->autoRepeat)
-        d->stopPressRepeat();
-    else if (d->holdTimer > 0 && (!d->pressed || QLineF(d->pressPoint, event->localPos()).length() > QGuiApplication::styleHints()->startDragDistance()))
-        d->stopPressAndHold();
-}
-
-void QQuickAbstractButton::mouseReleaseEvent(QMouseEvent *event)
-{
-    Q_D(QQuickAbstractButton);
-    QQuickControl::mouseReleaseEvent(event);
-    bool wasPressed = d->pressed;
-    setPressed(false);
-
-    if (!d->wasHeld && (d->keepPressed || contains(event->pos())))
-        nextCheckState();
-
-    if (wasPressed) {
-        emit released();
-        if (!d->wasHeld)
-            emit clicked();
-    } else {
-        emit canceled();
-    }
-
-    if (d->autoRepeat)
-        d->stopPressRepeat();
-    else
-        d->stopPressAndHold();
 }
 
 void QQuickAbstractButton::mouseDoubleClickEvent(QMouseEvent *event)
@@ -576,18 +630,6 @@ void QQuickAbstractButton::mouseDoubleClickEvent(QMouseEvent *event)
     emit doubleClicked();
 }
 
-void QQuickAbstractButton::mouseUngrabEvent()
-{
-    Q_D(QQuickAbstractButton);
-    QQuickControl::mouseUngrabEvent();
-    if (d->pressed) {
-        setPressed(false);
-        d->stopPressRepeat();
-        d->stopPressAndHold();
-        emit canceled();
-    }
-}
-
 void QQuickAbstractButton::timerEvent(QTimerEvent *event)
 {
     Q_D(QQuickAbstractButton);
@@ -605,13 +647,19 @@ void QQuickAbstractButton::timerEvent(QTimerEvent *event)
     }
 }
 
-void QQuickAbstractButton::checkStateSet()
+void QQuickAbstractButton::buttonChange(ButtonChange change)
 {
     Q_D(QQuickAbstractButton);
-    if (d->checked) {
-        QQuickAbstractButton *button = d->findCheckedButton();
-        if (button && button != this)
-            button->setChecked(false);
+    switch (change) {
+    case ButtonCheckedChange:
+        if (d->checked) {
+            QQuickAbstractButton *button = d->findCheckedButton();
+            if (button && button != this)
+                button->setChecked(false);
+        }
+        break;
+    default:
+        break;
     }
 }
 
@@ -619,18 +667,10 @@ void QQuickAbstractButton::nextCheckState()
 {
     Q_D(QQuickAbstractButton);
     if (d->checkable && (!d->checked || d->findCheckedButton() != this))
-        setChecked(!d->checked);
+        d->toggle(!d->checked);
 }
 
-void QQuickAbstractButton::checkableChange()
-{
-}
-
-void QQuickAbstractButton::autoRepeatChange()
-{
-}
-
-#ifndef QT_NO_ACCESSIBILITY
+#if QT_CONFIG(accessibility)
 void QQuickAbstractButton::accessibilityActiveChanged(bool active)
 {
     QQuickControl::accessibilityActiveChanged(active);
