@@ -1,6 +1,6 @@
 /****************************************************************************
 **
-** Copyright (C) 2015 The Qt Company Ltd.
+** Copyright (C) 2017 The Qt Company Ltd.
 ** Contact: http://www.qt.io/licensing/
 **
 ** This file is part of the test suite of the Qt Toolkit.
@@ -44,6 +44,7 @@
 #include <QtGui/qguiapplication.h>
 #include <QtGui/qpa/qwindowsysteminterface.h>
 #include <QtQuick/private/qquickwindow_p.h>
+#include <QtQuick/private/qquickflickable_p.h>
 #include <QtQuickTemplates2/private/qquickapplicationwindow_p.h>
 #include <QtQuickTemplates2/private/qquickoverlay_p.h>
 #include <QtQuickTemplates2/private/qquickpopup_p_p.h>
@@ -58,6 +59,8 @@ class tst_Drawer : public QQmlDataTest
     Q_OBJECT
 
 private slots:
+    void initTestCase();
+
     void visible_data();
     void visible();
 
@@ -83,9 +86,46 @@ private slots:
     void touch_data();
     void touch();
 
+    void multiTouch();
+
     void grabber();
+
+    void interactive_data();
+    void interactive();
+
+    void flickable_data();
+    void flickable();
+
+    void dragOverModalShadow_data();
+    void dragOverModalShadow();
+
+    void nonModal_data();
+    void nonModal();
+
+private:
+    struct TouchDeviceDeleter
+    {
+        static inline void cleanup(QTouchDevice *device)
+        {
+            QWindowSystemInterface::unregisterTouchDevice(device);
+            delete device;
+        }
+    };
+
+    QScopedPointer<QTouchDevice, TouchDeviceDeleter> touchDevice;
 };
 
+
+void tst_Drawer::initTestCase()
+{
+    QQmlDataTest::initTestCase();
+    qputenv("QML_NO_TOUCH_COMPRESSION", "1");
+
+    touchDevice.reset(new QTouchDevice);
+    touchDevice->setType(QTouchDevice::TouchScreen);
+    QWindowSystemInterface::registerTouchDevice(touchDevice.data());
+}
+
 void tst_Drawer::visible_data()
 {
     QTest::addColumn<QString>("source");
@@ -237,19 +277,21 @@ void tst_Drawer::state()
 void tst_Drawer::position_data()
 {
     QTest::addColumn<Qt::Edge>("edge");
+    QTest::addColumn<QPoint>("press");
     QTest::addColumn<QPoint>("from");
     QTest::addColumn<QPoint>("to");
     QTest::addColumn<qreal>("position");
 
-    QTest::newRow("top") << Qt::TopEdge << QPoint(100, 0) << QPoint(100, 100) << qreal(0.5);
-    QTest::newRow("left") << Qt::LeftEdge << QPoint(0, 100) << QPoint(100, 100) << qreal(0.5);
-    QTest::newRow("right") << Qt::RightEdge << QPoint(399, 100) << QPoint(300, 100) << qreal(0.5);
-    QTest::newRow("bottom") << Qt::BottomEdge << QPoint(100, 399) << QPoint(100, 300) << qreal(0.5);
+    QTest::newRow("top") << Qt::TopEdge << QPoint(100, 0) << QPoint(100, 50) << QPoint(100, 150) << qreal(0.5);
+    QTest::newRow("left") << Qt::LeftEdge << QPoint(0, 100) << QPoint(50, 100) << QPoint(150, 100) << qreal(0.5);
+    QTest::newRow("right") << Qt::RightEdge << QPoint(399, 100) << QPoint(350, 100) << QPoint(250, 100) << qreal(0.5);
+    QTest::newRow("bottom") << Qt::BottomEdge << QPoint(100, 399) << QPoint(100, 350) << QPoint(150, 250) << qreal(0.5);
 }
 
 void tst_Drawer::position()
 {
     QFETCH(Qt::Edge, edge);
+    QFETCH(QPoint, press);
     QFETCH(QPoint, from);
     QFETCH(QPoint, to);
     QFETCH(qreal, position);
@@ -265,7 +307,8 @@ void tst_Drawer::position()
     QVERIFY(drawer);
     drawer->setEdge(edge);
 
-    QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, from);
+    QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, press);
+    QTest::mouseMove(window, from);
     QTest::mouseMove(window, to);
     QCOMPARE(drawer->position(), position);
 
@@ -311,26 +354,30 @@ void tst_Drawer::dragMargin()
     drawer->setEdge(edge);
     drawer->setDragMargin(dragMargin);
 
+    const int startDragDistance = qMax(20, QGuiApplication::styleHints()->startDragDistance() + 5) + 1;
+
     // drag from the left
     int leftX = qMax<int>(0, dragMargin);
-    int leftDistance = drawer->width() * 0.45;
+    int leftDistance = startDragDistance + drawer->width() * 0.45;
     QVERIFY(leftDistance > QGuiApplication::styleHints()->startDragDistance());
     QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, QPoint(leftX, drawer->height() / 2));
-    QTest::mouseMove(window, QPoint(leftDistance, drawer->height() / 2));
+    QTest::mouseMove(window, QPoint(leftX + startDragDistance, drawer->height() / 2));
+    QTest::mouseMove(window, QPoint(leftX + leftDistance, drawer->height() / 2));
     QCOMPARE(drawer->position(), dragFromLeft);
-    QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, QPoint(leftDistance, drawer->height() / 2));
+    QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, QPoint(leftX + leftDistance, drawer->height() / 2));
 
     drawer->close();
     QTRY_COMPARE(drawer->position(), qreal(0.0));
 
     // drag from the right
     int rightX = qMin<int>(window->width() - 1, window->width() - dragMargin);
-    int rightDistance = drawer->width() * 0.75;
+    int rightDistance = startDragDistance + drawer->width() * 0.75;
     QVERIFY(rightDistance > QGuiApplication::styleHints()->startDragDistance());
     QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, QPoint(rightX, drawer->height() / 2));
-    QTest::mouseMove(window, QPoint(window->width() - rightDistance, drawer->height() / 2));
+    QTest::mouseMove(window, QPoint(rightX - startDragDistance, drawer->height() / 2));
+    QTest::mouseMove(window, QPoint(rightX - rightDistance, drawer->height() / 2));
     QCOMPARE(drawer->position(), dragFromRight);
-    QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, QPoint(window->width() - rightDistance, drawer->height() / 2));
+    QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, QPoint(rightX - rightDistance, drawer->height() / 2));
 }
 
 static QRectF geometry(const QQuickItem *item)
@@ -380,10 +427,20 @@ void tst_Drawer::reposition()
 
     window->setWidth(window->width() + 100);
     QTRY_COMPARE(geometry(dimmer), QRectF(0, 150, window->width(), window->height() - 150));
-    QCOMPARE(geometry(popupItem), QRectF(window->width() - drawer->width(), 150, window->width() / 2, window->height() - 150));
+    QTRY_COMPARE(geometry(popupItem), QRectF(window->width() - drawer->width(), 150, window->width() / 2, window->height() - 150));
 
     drawer->close();
     QTRY_COMPARE(geometry(popupItem), QRectF(window->width(), 150, window->width() / 2, window->height() - 150));
+
+    QQuickDrawer *drawer2 = window->property("drawer2").value<QQuickDrawer *>();
+    QVERIFY(drawer2);
+    QQuickItem *popupItem2 = drawer2->popupItem();
+    QVERIFY(popupItem2);
+
+    drawer2->open();
+    QVERIFY(popupItem2->isVisible());
+    QCOMPARE(popupItem2->x(), -drawer2->width());
+    QTRY_COMPARE(popupItem2->x(), 0.0);
 }
 
 void tst_Drawer::header()
@@ -452,6 +509,10 @@ void tst_Drawer::hover()
     QVERIFY(drawer);
     drawer->setModal(modal);
 
+    QQuickControl *drawerItem = qobject_cast<QQuickControl *>(drawer->popupItem());
+    QVERIFY(drawerItem);
+    QVERIFY(drawerItem->isHoverEnabled());
+
     QQuickButton *backgroundButton = window->property("backgroundButton").value<QQuickButton*>();
     QVERIFY(backgroundButton);
     backgroundButton->setHoverEnabled(true);
@@ -469,16 +530,19 @@ void tst_Drawer::hover()
     QTest::mouseMove(window, QPoint(window->width() - 1, window->height() - 1));
     QCOMPARE(backgroundButton->isHovered(), !modal);
     QVERIFY(!drawerButton->isHovered());
+    QVERIFY(!drawerItem->isHovered());
 
     // hover the drawer background
     QTest::mouseMove(window, QPoint(1, 1));
     QVERIFY(!backgroundButton->isHovered());
     QVERIFY(!drawerButton->isHovered());
+    QVERIFY(drawerItem->isHovered());
 
     // hover the button in a drawer
     QTest::mouseMove(window, QPoint(2, 2));
     QVERIFY(!backgroundButton->isHovered());
     QVERIFY(drawerButton->isHovered());
+    QVERIFY(drawerItem->isHovered());
 
     QSignalSpy closedSpy(drawer, SIGNAL(closed()));
     QVERIFY(closedSpy.isValid());
@@ -488,6 +552,8 @@ void tst_Drawer::hover()
     // hover the background button after closing the drawer
     QTest::mouseMove(window, QPoint(window->width() / 2, window->height() / 2));
     QVERIFY(backgroundButton->isHovered());
+    QVERIFY(!drawerButton->isHovered());
+    QVERIFY(!drawerItem->isHovered());
 }
 
 void tst_Drawer::wheel_data()
@@ -608,7 +674,8 @@ void tst_Drawer::multiple()
 
     // drag the left drawer open
     QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, QPoint(0, window->height() / 2));
-    QTest::mouseMove(window, QPoint(leftDrawer->width() / 2, window->height() / 2));
+    QTest::mouseMove(window, QPoint(leftDrawer->width() / 4, window->height() / 2));
+    QTest::mouseMove(window, QPoint(leftDrawer->width() / 4 * 3, window->height() / 2));
     QCOMPARE(leftDrawer->position(), 0.5);
     QCOMPARE(rightDrawer->position(), 0.0);
     QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, QPoint(leftDrawer->width() / 2, window->height() / 2));
@@ -687,7 +754,8 @@ void tst_Drawer::multiple()
 
     // drag the right drawer open
     QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, QPoint(window->width() - 1, window->height() / 2));
-    QTest::mouseMove(window, QPoint(window->width() - rightDrawer->width() / 2, window->height() / 2));
+    QTest::mouseMove(window, QPoint(window->width() - rightDrawer->width() / 4, window->height() / 2));
+    QTest::mouseMove(window, QPoint(window->width() - rightDrawer->width() / 4 * 3, window->height() / 2));
     QCOMPARE(rightDrawer->position(), 0.5);
     QCOMPARE(leftDrawer->position(), 0.0);
     QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, QPoint(window->width() - rightDrawer->width() / 2, window->height() / 2));
@@ -698,13 +766,26 @@ void tst_Drawer::multiple()
 void tst_Drawer::touch_data()
 {
     QTest::addColumn<QString>("source");
-    QTest::newRow("Window") << "window.qml";
-    QTest::newRow("ApplicationWindow") << "applicationwindow.qml";
+    QTest::addColumn<QPoint>("from");
+    QTest::addColumn<QPoint>("to");
+
+    QTest::newRow("Window:inside") << "window.qml" << QPoint(150, 100) << QPoint(50, 100);
+    QTest::newRow("Window:outside") << "window.qml" << QPoint(300, 100) << QPoint(100, 100);
+    QTest::newRow("ApplicationWindow:inside") << "applicationwindow.qml" << QPoint(150, 100) << QPoint(50, 100);
+    QTest::newRow("ApplicationWindow:outside") << "applicationwindow.qml" << QPoint(300, 100) << QPoint(100, 100);
+
+    QTest::newRow("Window+Button:inside") << "window-button.qml" << QPoint(150, 100) << QPoint(50, 100);
+    QTest::newRow("Window+Button:outside") << "window-button.qml" << QPoint(300, 100) << QPoint(100, 100);
+    QTest::newRow("ApplicationWindow+Button:inside") << "applicationwindow-button.qml" << QPoint(150, 100) << QPoint(50, 100);
+    QTest::newRow("ApplicationWindow+Button:outside") << "applicationwindow-button.qml" << QPoint(300, 100) << QPoint(100, 100);
 }
 
 void tst_Drawer::touch()
 {
     QFETCH(QString, source);
+    QFETCH(QPoint, from);
+    QFETCH(QPoint, to);
+
     QQuickApplicationHelper helper(this, source);
 
     QQuickWindow *window = helper.window;
@@ -714,37 +795,140 @@ void tst_Drawer::touch()
     QQuickDrawer *drawer = window->property("drawer").value<QQuickDrawer*>();
     QVERIFY(drawer);
 
-    struct TouchDeviceDeleter
-    {
-        static inline void cleanup(QTouchDevice *device)
-        {
-            QWindowSystemInterface::unregisterTouchDevice(device);
-            delete device;
-        }
-    };
-
-    QScopedPointer<QTouchDevice, TouchDeviceDeleter> device(new QTouchDevice);
-    device->setType(QTouchDevice::TouchScreen);
-    QWindowSystemInterface::registerTouchDevice(device.data());
+    QSignalSpy drawerOpenedSpy(drawer, SIGNAL(opened()));
+    QSignalSpy drawerClosedSpy(drawer, SIGNAL(closed()));
+    QVERIFY(drawerOpenedSpy.isValid());
+    QVERIFY(drawerClosedSpy.isValid());
 
     // drag to open
-    QTest::touchEvent(window, device.data()).press(0, QPoint(0, 100));
-    QTest::touchEvent(window, device.data()).move(0, QPoint(100, 100));
-    QTRY_COMPARE(drawer->position(), 0.5);
-    QTest::touchEvent(window, device.data()).release(0, QPoint(100, 100));
-    QTRY_COMPARE(drawer->position(), 1.0);
+    QTest::touchEvent(window, touchDevice.data()).press(0, QPoint(0, 100));
+    QTest::touchEvent(window, touchDevice.data()).move(0, QPoint(50, 100));
+    QTest::touchEvent(window, touchDevice.data()).move(0, QPoint(150, 100));
+    QTest::touchEvent(window, touchDevice.data()).release(0, QPoint(150, 100));
+    QVERIFY(drawerOpenedSpy.wait());
+    QCOMPARE(drawer->position(), 1.0);
 
     // drag to close
-    QTest::touchEvent(window, device.data()).press(0, QPoint(300, 100));
-    QTest::touchEvent(window, device.data()).move(0, QPoint(300 - drawer->dragMargin(), 100));
-    for (int x = 300; x > 100; x -= 10) {
-        QTest::touchEvent(window, device.data()).move(0, QPoint(x, 100));
-        QQuickWindowPrivate::get(window)->flushDelayedTouchEvent();
+    QTest::touchEvent(window, touchDevice.data()).press(0, from);
+    for (int x = from.x(); x > to.x(); x -= 10)
+        QTest::touchEvent(window, touchDevice.data()).move(0, QPoint(x, to.y()));
+    QTest::touchEvent(window, touchDevice.data()).move(0, to);
+    QTest::touchEvent(window, touchDevice.data()).release(0, to);
+    QVERIFY(drawerClosedSpy.wait());
+    QCOMPARE(drawer->position(), 0.0);
+}
+
+void tst_Drawer::multiTouch()
+{
+    QQuickApplicationHelper helper(this, QStringLiteral("multiTouch.qml"));
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowActive(window));
+
+    QQuickOverlay *overlay = QQuickOverlay::overlay(window);
+    QVERIFY(overlay);
+
+    QQuickDrawer *drawer = window->property("drawer").value<QQuickDrawer *>();
+    QVERIFY(drawer);
+
+    QQuickPopup *popup = window->property("popup").value<QQuickPopup *>();
+    QVERIFY(popup);
+
+    QQuickButton *button = window->property("button").value<QQuickButton *>();
+    QVERIFY(button);
+
+    QSignalSpy overlayPressedSpy(overlay, SIGNAL(pressed()));
+    QSignalSpy overlayReleasedSpy(overlay, SIGNAL(released()));
+    QVERIFY(overlayPressedSpy.isValid());
+    QVERIFY(overlayReleasedSpy.isValid());
+
+    QSignalSpy drawerOpenedSpy(drawer, SIGNAL(opened()));
+    QVERIFY(drawerOpenedSpy.isValid());
+
+    QSignalSpy buttonPressedSpy(button, SIGNAL(pressed()));
+    QSignalSpy buttonReleasedSpy(button, SIGNAL(released()));
+    QVERIFY(buttonPressedSpy.isValid());
+    QVERIFY(buttonReleasedSpy.isValid());
+
+    popup->open();
+    QVERIFY(popup->isVisible());
+
+    drawer->open();
+    QVERIFY(drawer->isVisible());
+    QVERIFY(drawerOpenedSpy.wait());
+
+    // 1st press
+    QTest::touchEvent(window, touchDevice.data()).press(0, QPoint(300, 100));
+    QVERIFY(popup->isVisible());
+    QVERIFY(drawer->isVisible());
+    QCOMPARE(buttonPressedSpy.count(), 0);
+    QCOMPARE(overlayPressedSpy.count(), 1);
+
+    // 2nd press (blocked & ignored)
+    QTest::touchEvent(window, touchDevice.data()).stationary(0).press(1, QPoint(300, 200));
+    QVERIFY(popup->isVisible());
+    QVERIFY(drawer->isVisible());
+    QCOMPARE(buttonPressedSpy.count(), 0);
+    QCOMPARE(overlayPressedSpy.count(), 2);
+
+    // 2nd release (blocked & ignored)
+    QTest::touchEvent(window, touchDevice.data()).stationary(0).release(1, QPoint(300, 200));
+    QVERIFY(popup->isVisible());
+    QVERIFY(drawer->isVisible());
+    QCOMPARE(buttonPressedSpy.count(), 0);
+    QCOMPARE(buttonReleasedSpy.count(), 0);
+    QCOMPARE(overlayPressedSpy.count(), 2);
+    QCOMPARE(overlayReleasedSpy.count(), 1);
+
+    // 1st release
+    QTest::touchEvent(window, touchDevice.data()).release(0, QPoint(300, 100));
+    QVERIFY(popup->isVisible());
+    QTRY_VERIFY(!drawer->isVisible());
+    QCOMPARE(buttonPressedSpy.count(), 0);
+    QCOMPARE(buttonReleasedSpy.count(), 0);
+    QCOMPARE(overlayPressedSpy.count(), 2);
+    QCOMPARE(overlayReleasedSpy.count(), 2);
+
+    drawer->open();
+    QVERIFY(drawer->isVisible());
+    QVERIFY(drawerOpenedSpy.wait());
+
+    // 1st drag
+    QTest::touchEvent(window, touchDevice.data()).press(0, QPoint(300, 100));
+    QCOMPARE(buttonPressedSpy.count(), 0);
+    QCOMPARE(overlayPressedSpy.count(), 3);
+    for (int x = 300; x >= 100; x -= 10) {
+        QTest::touchEvent(window, touchDevice.data()).move(0, QPoint(x, 100));
+        QVERIFY(popup->isVisible());
+        QVERIFY(drawer->isVisible());
     }
-    QTest::touchEvent(window, device.data()).move(0, QPoint(100, 100));
-    QTRY_COMPARE(drawer->position(), 0.5);
-    QTest::touchEvent(window, device.data()).release(0, QPoint(100, 100));
-    QTRY_COMPARE(drawer->position(), 0.0);
+    QCOMPARE(drawer->position(), 0.5);
+
+    // 2nd drag (blocked & ignored)
+    QTest::touchEvent(window, touchDevice.data()).stationary(0).press(1, QPoint(300, 200));
+    QCOMPARE(buttonPressedSpy.count(), 0);
+    QCOMPARE(overlayPressedSpy.count(), 4);
+    for (int x = 300; x >= 0; x -= 10) {
+        QTest::touchEvent(window, touchDevice.data()).stationary(0).move(1, QPoint(x, 200));
+        QVERIFY(popup->isVisible());
+        QVERIFY(drawer->isVisible());
+    }
+    QCOMPARE(drawer->position(), 0.5);
+
+    // 2nd release (blocked & ignored)
+    QTest::touchEvent(window, touchDevice.data()).stationary(0).release(1, QPoint(300, 0));
+    QVERIFY(popup->isVisible());
+    QVERIFY(drawer->isVisible());
+    QCOMPARE(drawer->position(), 0.5);
+    QCOMPARE(buttonReleasedSpy.count(), 0);
+    QCOMPARE(overlayReleasedSpy.count(), 3);
+
+    // 1st release
+    QTest::touchEvent(window, touchDevice.data()).release(0, QPoint(300, 100));
+    QVERIFY(popup->isVisible());
+    QTRY_VERIFY(!drawer->isVisible());
+    QCOMPARE(buttonReleasedSpy.count(), 0);
+    QCOMPARE(overlayReleasedSpy.count(), 4);
 }
 
 void tst_Drawer::grabber()
@@ -783,6 +967,258 @@ void tst_Drawer::grabber()
     QTRY_COMPARE(popupClosedSpy.count(), 1);
 }
 
+void tst_Drawer::interactive_data()
+{
+    QTest::addColumn<QString>("source");
+    QTest::newRow("Window") << "window.qml";
+    QTest::newRow("ApplicationWindow") << "applicationwindow.qml";
+}
+
+void tst_Drawer::interactive()
+{
+    QFETCH(QString, source);
+    QQuickApplicationHelper helper(this, source);
+
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowActive(window));
+
+    QQuickDrawer *drawer = window->property("drawer").value<QQuickDrawer*>();
+    QVERIFY(drawer);
+
+    drawer->setInteractive(false);
+
+    QSignalSpy openedSpy(drawer, SIGNAL(opened()));
+    QSignalSpy aboutToHideSpy(drawer, SIGNAL(aboutToHide()));
+    QVERIFY(openedSpy.isValid());
+    QVERIFY(aboutToHideSpy.isValid());
+
+    drawer->open();
+    QVERIFY(openedSpy.wait());
+
+    // click outside
+    QTest::mouseClick(window, Qt::LeftButton, Qt::NoModifier, QPoint(300, 100));
+    QCOMPARE(aboutToHideSpy.count(), 0);
+
+    // drag inside
+    QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, QPoint(drawer->width(), 0));
+    QTest::mouseMove(window, QPoint(0, 0));
+    QCOMPARE(drawer->position(), 1.0);
+    QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, QPoint(0, 0));
+    QCOMPARE(drawer->position(), 1.0);
+    QCOMPARE(aboutToHideSpy.count(), 0);
+
+    // drag outside
+    QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, QPoint(window->width() - 1, 0));
+    QTest::mouseMove(window, QPoint(0, 0));
+    QCOMPARE(drawer->position(), 1.0);
+    QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, QPoint(0, 0));
+    QCOMPARE(drawer->position(), 1.0);
+    QCOMPARE(aboutToHideSpy.count(), 0);
+
+    // close on escape
+    QTest::keyClick(window, Qt::Key_Escape);
+    QCOMPARE(aboutToHideSpy.count(), 0);
+}
+
+void tst_Drawer::flickable_data()
+{
+    QTest::addColumn<bool>("mouse");
+    QTest::addColumn<QPoint>("from");
+    QTest::addColumn<QPoint>("to");
+
+    QTest::newRow("mouse,straight") << true << QPoint(200, 200) << QPoint(200, 100);
+    QTest::newRow("mouse,diagonal") << true << QPoint(200, 200) << QPoint(250, 100);
+    QTest::newRow("touch,straight") << false << QPoint(200, 200) << QPoint(200, 100);
+    QTest::newRow("touch,diagonal") << false << QPoint(200, 200) << QPoint(250, 100);
+}
+
+void tst_Drawer::flickable()
+{
+    QFETCH(bool, mouse);
+    QFETCH(QPoint, from);
+    QFETCH(QPoint, to);
+
+    QQuickApplicationHelper helper(this, QStringLiteral("flickable.qml"));
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowExposed(window));
+
+    QQuickDrawer *drawer = window->property("drawer").value<QQuickDrawer *>();
+    QVERIFY(drawer);
+
+    QQuickFlickable *flickable = window->property("flickable").value<QQuickFlickable *>();
+    QVERIFY(flickable);
+
+    QSignalSpy drawerOpenedSpy(drawer, SIGNAL(opened()));
+    QVERIFY(drawerOpenedSpy.isValid());
+
+    drawer->open();
+    QVERIFY(drawerOpenedSpy.wait());
+
+    if (mouse)
+        QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, from);
+    else
+        QTest::touchEvent(window, touchDevice.data()).press(0, from);
+
+    static const int steps = 10;
+    for (int i = 0; i < steps; ++i) {
+        int x = i * qAbs(from.x() - to.x()) / steps;
+        int y = i * qAbs(from.y() - to.y()) / steps;
+
+        if (mouse)
+            QTest::mouseMove(window, QPoint(x, y));
+        else
+            QTest::touchEvent(window, touchDevice.data()).move(0, QPoint(x, y));
+        QTest::qWait(1); // avoid infinite velocity
+    }
+
+    QVERIFY(flickable->isDragging());
+
+    if (mouse)
+        QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, to);
+    else
+        QTest::touchEvent(window, touchDevice.data()).release(0, to);
+
+    QVERIFY(!flickable->isDragging());
+}
+
+void tst_Drawer::dragOverModalShadow_data()
+{
+    QTest::addColumn<bool>("mouse");
+    QTest::newRow("mouse") << true;
+    QTest::newRow("touch") << false;
+}
+
+// QTBUG-60602
+void tst_Drawer::dragOverModalShadow()
+{
+    QFETCH(bool, mouse);
+
+    QQuickApplicationHelper helper(this, QStringLiteral("dragOverModalShadow.qml"));
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowActive(window));
+
+    QQuickDrawer *drawer = window->property("drawer").value<QQuickDrawer *>();
+    QVERIFY(drawer);
+
+    QQuickPopup *popup = window->property("popup").value<QQuickPopup *>();
+    QVERIFY(popup);
+
+    popup->open();
+    QVERIFY(popup->isVisible());
+    QVERIFY(!drawer->isVisible());
+
+    const QPoint from(popup->x(), popup->y() + popup->height() + 5);
+    const QPoint to(popup->x() + popup->width(), popup->y() + popup->height() + 5);
+
+    if (mouse)
+        QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, from);
+    else
+        QTest::touchEvent(window, touchDevice.data()).press(0, from);
+    QVERIFY(!drawer->isVisible());
+
+    static const int steps = 10;
+    for (int i = 0; i < steps; ++i) {
+        int x = from.x() + i * qAbs(from.x() - to.x()) / steps;
+        int y = from.y() + i * qAbs(from.y() - to.y()) / steps;
+
+        if (mouse)
+            QTest::mouseMove(window, QPoint(x, y));
+        else
+            QTest::touchEvent(window, touchDevice.data()).move(0, QPoint(x, y));
+        QTest::qWait(1); // avoid infinite velocity
+        QVERIFY(!drawer->isVisible());
+    }
+
+    if (mouse)
+        QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, to);
+    else
+        QTest::touchEvent(window, touchDevice.data()).release(0, to);
+    QVERIFY(!drawer->isVisible());
+}
+
+void tst_Drawer::nonModal_data()
+{
+    QTest::addColumn<bool>("mouse");
+    QTest::newRow("mouse") << true;
+    QTest::newRow("touch") << false;
+}
+
+// QTBUG-59652
+void tst_Drawer::nonModal()
+{
+    QFETCH(bool, mouse);
+
+    QQuickApplicationHelper helper(this, QStringLiteral("window.qml"));
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowActive(window));
+
+    QQuickDrawer *drawer = window->property("drawer").value<QQuickDrawer *>();
+    QVERIFY(drawer);
+    drawer->setModal(false);
+
+    const QPoint from(1, 1);
+    const QPoint to(150, 1);
+
+    // drag to open
+    QSignalSpy openedSpy(drawer, SIGNAL(opened()));
+    QVERIFY(openedSpy.isValid());
+
+    if (mouse)
+        QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, from);
+    else
+        QTest::touchEvent(window, touchDevice.data()).press(0, from);
+
+    static const int steps = 10;
+    for (int i = 0; i < steps; ++i) {
+        int x = i * qAbs(from.x() - to.x()) / steps;
+        int y = i * qAbs(from.y() - to.y()) / steps;
+
+        if (mouse)
+            QTest::mouseMove(window, QPoint(x, y));
+        else
+            QTest::touchEvent(window, touchDevice.data()).move(0, QPoint(x, y));
+        QTest::qWait(1); // avoid infinite velocity
+    }
+    QVERIFY(drawer->isVisible());
+
+    if (mouse)
+        QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, to);
+    else
+        QTest::touchEvent(window, touchDevice.data()).release(0, to);
+    QVERIFY(openedSpy.wait());
+
+    // drag to close
+    QSignalSpy closedSpy(drawer, SIGNAL(closed()));
+    QVERIFY(closedSpy.isValid());
+
+    if (mouse)
+        QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, to);
+    else
+        QTest::touchEvent(window, touchDevice.data()).press(0, to);
+
+    for (int i = steps - 1; i >= 0; --i) {
+        int x = i * qAbs(from.x() - to.x()) / steps;
+        int y = i * qAbs(from.y() - to.y()) / steps;
+
+        if (mouse)
+            QTest::mouseMove(window, QPoint(x, y));
+        else
+            QTest::touchEvent(window, touchDevice.data()).move(0, QPoint(x, y));
+        QTest::qWait(1); // avoid infinite velocity
+    }
+    QVERIFY(drawer->isVisible());
+
+    if (mouse)
+        QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, from);
+    else
+        QTest::touchEvent(window, touchDevice.data()).release(0, from);
+    QVERIFY(closedSpy.wait());
+}
+
 QTEST_MAIN(tst_Drawer)
 
 #include "tst_drawer.moc"
